/*
 * This file is part of JPuzzle.
 *
 * JPuzzle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * JPuzzle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JPuzzle.  If not, see <http://www.gnu.org/licenses/>.
 *
 * JPuzzle Copyright © 2008 HackmanGT
 * HackmanGT@hotmail.com
 */

package JPuzzle.Utilities;

import JPuzzle.Graphics.Matriz;
import java.io.IOException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.EventListenerList;

/**
 * El tablero es el controlador para realizar el juego propiamente, se encarga
 * de todas las operaciones de creación, modificación, y actualización del
 * modelo de datos, haciendo operaciones con todas las celdas en la tabla.
 * Maneja la matriz de imágenes que se utilizan en las celdas.
 * No es multi-threading!
 * @author HackmanGT
 */
public class Tablero extends Tabla {

    /**
     * El objeto generador de imágenes.
     */
    protected Matriz matriz = null;
    /**
     * El evento enviado cuando cambia el modelo de datos.
     */
    protected ChangeEvent changeEvent = null;
    /**
     * La lista de vistas que reciben eventos.
     */
    protected EventListenerList listenerList = new EventListenerList();

    /**
     * Construye una nuevo tablero con el tamaño por defecto de 4 x 4.
     */
    public Tablero() {
        initComponents(4, 4);
    }

    /**
     * Construye un nuevo tablero con los parámetros indicados en ancho y alto.
     * @param ancho el valor inicial del ancho de la tabla
     * @param alto el valor inicial del alto de la tabla
     */
    public Tablero(int ancho, int alto) {
        initComponents(ancho, alto);
    }

    /**
     * Inicializa el componente y carga las imágenes.
     * @param ancho el valor inicial del ancho de la tabla
     * @param alto el valor inicial del alto de la tabla
     */
    private void initComponents(int ancho, int alto) {
        this.setAncho(ancho);
        this.setAlto(alto);
        this.matriz = new Matriz();
        try {
            this.matriz.loadImage();
        } catch (IOException ex) {
            System.err.println(ex.getMessage());
        }
        crear();
    }

    /**
     * Intercambia una celda especificada en la posición origen, por otra que
     * se indica por la posición destino en la tabla de modelo de datos.
     * Y avisa a la vista que cambiaron los datos.
     * @param origen la posición de la celda que se debe intercambiar
     * @param destino la posición de la celda con la cual debe de intercambiarse
     */
    protected void intercambiarCelda(Posicion origen, Posicion destino) {
        Celda c = celdas[destino.getX()][destino.getY()];
        celdas[destino.getX()][destino.getY()] = celdas[origen.getX()][origen.getY()];
        celdas[origen.getX()][origen.getY()] = c;
        fireStateChanged();
    }

    /**
     * Crea la tabla de modelo de datos, asignandole un valor consecutivo
     * incremental a cada celda, así como una parte de la imagen a mostrar.
     * La última celda la elimina, poniendole el valor nulo para indicar vacío.
     */
    protected void crear() {
        int valor = 1;
        celdas = new Celda[ancho][alto];
        for (int y = 0; y < alto; y++) {
            for (int x = 0; x < ancho; x++) {
                celdas[x][y] = new Celda(valor++);
                celdas[x][y].setImagen(matriz.getTile(x, y, ancho, alto));
            }
        }
        celdas[ancho - 1][alto - 1] = null;
    }

    /**
     * Intenta encontrar la celda vacía en la tabla, indicada por el valor nulo.
     * @return la posición de la celda que está vacía, o nulo si no la encuentra
     */
    protected Posicion encontrarVacio() {
        for (int y = 0; y < alto; y++) {
            for (int x = 0; x < ancho; x++) {
                if (celdas[x][y] == null) {
                    return new Posicion(x, y);
                }
            }
        }
        return null;
    }

    /**
     * Intenta encontrar una celda en la tabla, indicada por su valor.
     * @param valor el número ordinal de la celda en la tabla
     * @return la posición de la celda buscada, o nulo si no la encuentra
     */
    protected Posicion encontrarCelda(int valor) {
        for (int y = 0; y < alto; y++) {
            for (int x = 0; x < ancho; x++) {
                if (celdas[x][y] != null) {
                    if (celdas[x][y].getValor() == valor) {
                        return new Posicion(x, y);
                    }
                }
            }
        }
        return null;
    }

    /**
     * Intenta realizar un movimiento dado por la posicion p, validando todos
     * los posibles movimimientos cercanos, y si es válido realiza la operación
     * de intercambiar por el espacio vacío.
     * @param p la posición que se desea mover al espacio vacío
     * @return verdadero si se realizo el movimiento, sino falso
     */
    public boolean intentarMover(Posicion p) {
        Posicion n = null;
        for (Posicion.Direccion d : Posicion.Direccion.values()) {
            n = p.relativa(d);
            if (esPosicionValida(n)) {
                if (celdas[n.getX()][n.getY()] == null) {
                    intercambiarCelda(n, p);
                    return true;
                }
            }
        }
        return false;
    }

    /**
     * Verifica si todas las celdas están ordenadas naturalmente. Es decir,
     * verifica si el jugador ha logrado ordenar las celdas.
     * @return verdadero si estan ordenadas, sino falso
     */
    public boolean verificarGanador() {
        int valor = 1;
        for (int y = 0; y < alto; y++) {
            for (int x = 0; x < ancho; x++) {
                if (celdas[x][y] != null) {
                    if (celdas[x][y].getValor() != valor) {
                        return false;
                    }
                }
                valor++;
            }
        }
        return true;
    }

    /**
     * Muestra una vista previa del modelo de datos, en la consola de Java.
     * Tomando en cuenta los saltos de línea para mantener el formato.
     * (Depuración)
     */
    public void preview() {
        for (int y = 0; y < alto; y++) {
            for (int x = 0; x < ancho; x++) {
                System.out.format("%2d ",
                        celdas[x][y] != null ?
                        celdas[x][y].getValor() :
                        -1);
            }
            System.out.println();
        }
    }

    /**
     * Desordena la tabla realizando movimientos aleatorios de cualquiera
     * celda que se encuentre en las cuatro direcciónes cercana a la celda
     * vacía, el objetivo es simular el (des)ordenamiento natural humano.
     * Intenta no regresar a la celda de donde viene.
     * @param nivel la cantidad de movimientos aleatorios a realizar
     */
    public void desordenar(int nivel) {
        Posicion p = encontrarVacio();
        Posicion n = null;
        Posicion m = null;

        int s = 0;
        while (s < nivel) {
            n = p.relativa(Posicion.Direccion.random());
            if (!n.equals(m)) {
                if (esPosicionValida(n)) {
                    m = p;
                    intercambiarCelda(n, p);
                    p = n;
                    s++;
                }
            }
        }
    }

    /**
     * Resuelve el tablero.
     * Ver <b>Class Cerebro</b>
     */
    public void resolver() {
        Cerebro c = new Cerebro(this);
        c.Resolver();
    }

    /**
     * Agrega una vista receptora de eventos a la lista.
     * @param l la vista que va a recibir el evento
     */
    public void addChangeListener(ChangeListener l) {
        listenerList.add(ChangeListener.class, l);
    }

    /**
     * Remueve una vista receptora de eventos de la lista.
     * @param l la vista que se va a eliminar
     */
    public void removeChangeListener(ChangeListener l) {
        listenerList.remove(ChangeListener.class, l);
    }

    /**
     * Avisa a todas las vistas que hubo un cambio en la tabla modelo de datos.
     */
    protected void fireStateChanged() {
        Object[] listeners = listenerList.getListenerList();
        for (int i = listeners.length - 2; i >= 0; i -= 2) {
            if (listeners[i] == ChangeListener.class) {
                if (changeEvent == null) {
                    changeEvent = new ChangeEvent(this);
                }
                ((ChangeListener) listeners[i + 1]).stateChanged(changeEvent);
            }
        }
    }
}
